import	os
import	unittest
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.spells				import flagrate
from	pygwarts.hedwig.mail.letter			import ObjectField
from	pygwarts.hedwig.mail.letter			import StringField
from	pygwarts.hedwig.mail.letter			import AddressField
from	pygwarts.hedwig.mail.letter			import TextField
from	pygwarts.hedwig.mail.letter.fields	import SenderField
from	pygwarts.hedwig.mail.letter.fields	import RecipientField
from	pygwarts.hedwig.mail.letter.fields	import ccField
from	pygwarts.hedwig.mail.letter.fields	import bccField
from	pygwarts.hedwig.mail.letter.fields	import AttachmentField
from	pygwarts.hedwig.mail.letter.fields	import SubjectField
from	pygwarts.hedwig.mail.letter.fields	import BodyField
from	pygwarts.hedwig.mail.utils			import EmailValidator
from	pygwarts.hedwig.mail.builder		import LetterBuilder
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.tests.hedwig				import HedwigTestCase








class LetterFieldsCases(HedwigTestCase):


	SAMPLE_TEXT_1	= "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA\nBING BANG"
	SAMPLE_TEXT_2	= """OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA\nBING BANG"""
	SAMPLE_TEXT_3	= "O"
	SAMPLE_TEXT_4	= """O"""
	SAMPLE_TEXT_5	= ""
	SAMPLE_TEXT_6	= """"""
	SAMPLE_TEXT_7	= "{first} EEH\n{first} AH AH\nTING TANG\n{second} {second}\nBING BANG"
	SAMPLE_TEXT_8	= """{first} EEH\n{first} AH AH\nTING TANG\n{second} {second}\nBING BANG"""
	SAMPLE_MODIFS	= { "first": "OOH", "second": "WALLA" }


	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.HEDWIG_MAIL_LETTER_HANDLER):

				cls.current_text.loggy.close()
				cls.current_address.loggy.close()
				os.remove(cls.HEDWIG_MAIL_LETTER_HANDLER)


	@classmethod
	def setUpClass(cls):

		class CurrentAddress(AddressField):

			class validator(EmailValidator): pass
			class loggy(LibraryContrib):

				handler		= cls.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "CurrentAddress"
				init_level	= 10
		cls.current_address	= CurrentAddress()


		class CurrentText(TextField):
			class loggy(LibraryContrib):

				handler		= cls.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "CurrentText"
				init_level	= 10
		cls.current_text	= CurrentText()
		cls.make_loggy_file(cls, cls.HEDWIG_MAIL_LETTER_HANDLER)




	def test_ObjectField_values(self):

		for i,value in enumerate((

			1, 1., "1", True, False, None, print,
			[ "information" ],
			( "information", ),
			{ "information" },
			{ "field_value": "information" },
		)):
			with self.subTest(value=value):
				class CurrentField(ObjectField):

					field_value = value
					class loggy(LibraryContrib):

						handler		= self.HEDWIG_MAIL_LETTER_HANDLER
						init_name	= f"ObjectField_values_{i}"
						init_level	= 10

				with self.assertLogs(f"ObjectField_values_{i}", 10) as case_loggy:

					self.test_case	= CurrentField()
					current_value	= self.test_case()

				self.test_case.loggy.close()
				self.assertEqual(current_value, value)
				self.assertEqual(len(case_loggy.output),1)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(

					f"DEBUG:ObjectField_values_{i}:Fetched value of {type(value)}",
					case_loggy.output
				)
				self.test_case.loggy.close()




	def test_ObjectField_novalue(self):

		class CurrentField(ObjectField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= f"ObjectField_novalue"

		with self.assertLogs(f"ObjectField_novalue", 10) as case_loggy:

			self.test_case	= CurrentField()
			current_value	= self.test_case()

		self.assertIn(f"ERROR:ObjectField_novalue:Field {self.test_case} must have value",case_loggy.output)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertEqual(len(case_loggy.output),1)
		self.assertIsNone(current_value)








	def test_StringField_valid(self):
		class CurrentField(StringField):

			field_value = "hedwig's letter"
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= f"StringField_valid"
				init_level	= 10

		with self.assertLogs(f"StringField_valid", 10) as case_loggy:

			self.test_case	= CurrentField()
			current_value	= self.test_case()

		self.assertIn(

			f"DEBUG:StringField_valid:Fetched value of {type('')}",
			case_loggy.output
		)
		self.assertIn("DEBUG:StringField_valid:Fetched 15 symbols",case_loggy.output)
		self.assertEqual(current_value, "hedwig's letter")
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),2)




	def test_StringField_invalid(self):

		for i,value in enumerate((

			1, 1., True, False, print,
			[ "information" ],
			( "information", ),
			{ "information" },
			{ "field_value": "information" },
		)):
			with self.subTest(value=value):
				class CurrentField(StringField):

					field_value = value
					class loggy(LibraryContrib):

						handler		= self.HEDWIG_MAIL_LETTER_HANDLER
						init_name	= f"StringField_invalid_{i}"
						init_level	= 10

				with self.assertLogs(f"StringField_invalid_{i}", 10) as case_loggy:

					self.test_case	= CurrentField()
					current_value	= self.test_case()

				self.test_case.loggy.close()
				self.no_loggy_levels(case_loggy.output, 50)
				self.assertEqual(len(case_loggy.output),2)
				self.assertIsNone(current_value)
				self.assertIn(

					f"DEBUG:StringField_invalid_{i}:Fetched value of {type(value)}",
					case_loggy.output
				)
				self.assertIn(

					f"ERROR:StringField_invalid_{i}:Field {self.test_case} value must be string",
					case_loggy.output
				)
				self.test_case.loggy.close()




	def test_StringField_novalue(self):

		class CurrentField(StringField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= f"StringField_novalue"
				init_level	= 10

		with self.assertLogs(f"StringField_novalue", 10) as case_loggy:

			self.test_case	= CurrentField()
			current_value	= self.test_case()

		self.assertIn(f"ERROR:StringField_novalue:Field {self.test_case} must have value",case_loggy.output)
		self.assertIn("DEBUG:StringField_novalue:No value fetched for field",case_loggy.output)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertEqual(len(case_loggy.output),2)
		self.assertIsNone(current_value)




	def test_StringField_None(self):
		class CurrentField(StringField):

			field_value = None
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= f"StringField_None"
				init_level	= 10

		with self.assertLogs(f"StringField_None", 10) as case_loggy:

			self.test_case	= CurrentField()
			current_value	= self.test_case()

		self.assertIn(f"DEBUG:StringField_None:Fetched value of {type(None)}",case_loggy.output)
		self.assertIn("DEBUG:StringField_None:No value fetched for field",case_loggy.output)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),2)
		self.assertIsNone(current_value)








	def test_AddressField_single_valid(self):

		for address in self.VALID_EMAILS:
			with self.subTest(email=address):

				self.current_address.field_value = address
				with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

				self.assertEqual(current, f"{address};")
				self.assertEqual(len(case_loggy.output),3)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address}\" validated", case_loggy.output
				)
				self.assertIn(
					f"INFO:CurrentAddress:{self.current_address} obtained 1 email", case_loggy.output
				)


	def test_AddressField_single_valid_semicol(self):

		for address in ( f"{E};" for E in self.VALID_EMAILS ):
			with self.subTest(email=address):

				self.current_address.field_value = address
				with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

				self.assertEqual(current, address)
				self.assertEqual(len(case_loggy.output),3)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address[:-1]}\" validated", case_loggy.output
				)
				self.assertIn(
					f"INFO:CurrentAddress:{self.current_address} obtained 1 email", case_loggy.output
				)


	def test_AddressField_single_invalid_str(self):

		# shrinked to not include ; and space addresses
		for address in self.INVALID_EMAILS[1:-3]:
			with self.subTest(email=address):

				self.current_address.field_value = address
				with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

				self.assertIsNone(current)
				self.assertEqual(len(case_loggy.output),3)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address}\"", case_loggy.output
				)
				self.assertIn(

					f"INFO:CurrentAddress:{self.current_address} field has no valid addresses",
					case_loggy.output
				)


	def test_AddressField_single_invalid_str_semicol(self):

		# shrinked to not include semicolumned, spaced and empty addresses
		for address in ( f"{E};" for E in self.INVALID_EMAILS[1:-3] ):
			with self.subTest(email=address):

				self.current_address.field_value = address
				with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

				self.assertIsNone(current)
				self.assertEqual(len(case_loggy.output),3)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address[:-1]}\"", case_loggy.output
				)
				self.assertIn(

					f"INFO:CurrentAddress:{self.current_address} field has no valid addresses",
					case_loggy.output
				)


	def test_AddressField_joint_valid(self):

		joint_count = len(self.VALID_EMAILS)

		for valid_joint in (

			";".join(self.VALID_EMAILS),
			"; ".join(self.VALID_EMAILS),
			";	 ".join(self.VALID_EMAILS),
			"; 	".join(self.VALID_EMAILS),
			" ;".join(self.VALID_EMAILS),
			"	 ;".join(self.VALID_EMAILS),
			" 	;".join(self.VALID_EMAILS),
			" ; ".join(self.VALID_EMAILS),
			"	;".join(self.VALID_EMAILS),
			";	".join(self.VALID_EMAILS),
			"	;	".join(self.VALID_EMAILS),
		):
			self.current_address.field_value = valid_joint
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.assertEqual(len(case_loggy.output),joint_count +2)
			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} obtained {joint_count} emails",
				case_loggy.output
			)

			for address in self.VALID_EMAILS:
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address}\" validated", case_loggy.output
				)


	def test_AddressField_joint_invalid(self):

		# shrinked to not include semicolumned, spaced and empty addresses
		joint_source = self.INVALID_EMAILS[1:-3]

		for invalid_joint in (

			";".join(joint_source),
			"; ".join(joint_source),
			";	 ".join(joint_source),
			"; 	".join(joint_source),
			" ;".join(joint_source),
			"	 ;".join(joint_source),
			" 	;".join(joint_source),
			" ; ".join(joint_source),
			"	;".join(joint_source),
			";	".join(joint_source),
			"	;	".join(joint_source),
		):
			self.current_address.field_value = invalid_joint
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.assertEqual(len(case_loggy.output),len(joint_source) +2)
			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} field has no valid addresses",
				case_loggy.output
			)

			for address in joint_source:
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address}\"", case_loggy.output
				)


	def test_AddressField_joint_mixed(self):

		# shrinked to not include semicolumned, spaced and empty addresses
		invalid_joint_source = self.INVALID_EMAILS[1:-3]
		valid_count = len(self.VALID_EMAILS)
		joint_source = list(set(self.VALID_EMAILS) | set(invalid_joint_source))

		for mixed_joint in (

			";".join(joint_source),
			"; ".join(joint_source),
			";	 ".join(joint_source),
			"; 	".join(joint_source),
			" ;".join(joint_source),
			"	 ;".join(joint_source),
			" 	;".join(joint_source),
			" ; ".join(joint_source),
			"	;".join(joint_source),
			";	".join(joint_source),
			"	;	".join(joint_source),
		):
			self.current_address.field_value = mixed_joint
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),len(joint_source) +2)
			self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} obtained {valid_count} emails",
				case_loggy.output
			)

			for address in self.VALID_EMAILS:
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address}\" validated", case_loggy.output
				)
			for address in invalid_joint_source:
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address}\"", case_loggy.output
				)


	def test_AddressField_container_valid(self):

		joint_count = len(self.VALID_EMAILS)

		for valid_container in (

			self.VALID_EMAILS,
			list(self.VALID_EMAILS),
			set(self.VALID_EMAILS),
		):
			self.current_address.field_value = valid_container
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),joint_count +2)
			self.assertIn(
				f"DEBUG:CurrentAddress:Fetched value of {type(valid_container)}", case_loggy.output
			)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} obtained {joint_count} emails",
				case_loggy.output
			)

			for address in self.VALID_EMAILS:
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address}\" validated", case_loggy.output
				)


	def test_AddressField_container_invalid(self):

		# shrinked to not include semicolumned, spaced and empty addresses
		container_source = self.INVALID_EMAILS[1:-3]

		for invalid_container in (

			container_source,
			list(container_source),
			set(container_source),
		):
			self.current_address.field_value = invalid_container
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),len(container_source) +2)
			self.assertIn(
				f"DEBUG:CurrentAddress:Fetched value of {type(invalid_container)}", case_loggy.output
			)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} field has no valid addresses",
				case_loggy.output
			)

			for address in container_source:
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address}\"", case_loggy.output
				)


	def test_AddressField_continaer_mixed(self):

		# shrinked to not include semicolumned, spaced and empty addresses
		invalid_container_source = self.INVALID_EMAILS[1:-3]
		valid_count = len(self.VALID_EMAILS)
		container_source = tuple(set(self.VALID_EMAILS) | set(invalid_container_source))

		for mixed_container in (

			container_source,
			list(container_source),
			set(container_source),
		):
			self.current_address.field_value = mixed_container
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),len(container_source) +2)
			self.assertIn(
				f"DEBUG:CurrentAddress:Fetched value of {type(mixed_container)}", case_loggy.output
			)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} obtained {valid_count} emails",
				case_loggy.output
			)

			for address in self.VALID_EMAILS:
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address}\" validated", case_loggy.output
				)
			for address in invalid_container_source:
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address}\"", case_loggy.output
				)


	def test_AddressField_continaer_others_mixed(self):

		invalid_container_source = ( 1, False, print, LibraryContrib, )
		valid_count = len(self.VALID_EMAILS)
		container_source = tuple(set(self.VALID_EMAILS) | set(invalid_container_source))

		for mixed_container in (

			container_source,
			list(container_source),
			set(container_source),
		):
			self.current_address.field_value = mixed_container
			with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),len(container_source) +2)
			self.assertIn(
				f"DEBUG:CurrentAddress:Fetched value of {type(mixed_container)}", case_loggy.output
			)
			self.assertIn(

				f"INFO:CurrentAddress:{self.current_address} obtained {valid_count} emails",
				case_loggy.output
			)

			for address in self.VALID_EMAILS:
				self.assertIn(
					f"DEBUG:CurrentAddress:Email address \"{address}\" validated", case_loggy.output
				)
			for address in invalid_container_source:
				self.assertIn(
					f"INFO:CurrentAddress:Invalid email address \"{address}\"", case_loggy.output
				)


	def test_AddressField_invalid_field(self):

		for invalid_value in ( 1, 1., True, False, print, LibraryContrib, ):
			with self.subTest(value=invalid_value):
				self.current_address.field_value = invalid_value

				with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()

				self.no_loggy_levels(case_loggy.output, 50)
				self.assertEqual(len(case_loggy.output),2)
				self.assertIsNone(current)
				self.assertIn(
					f"DEBUG:CurrentAddress:Fetched value of {type(invalid_value)}", case_loggy.output
				)
				self.assertIn(

					f"ERROR:CurrentAddress:{self.current_address} field failed due to "
					f"TypeError: '{invalid_value.__class__.__name__}' object is not iterable",
					case_loggy.output
				)

		self.current_address.field_value = None
		with self.assertLogs("CurrentAddress", 10) as case_loggy : current = self.current_address()
		self.assertIn(f"DEBUG:CurrentAddress:Fetched value of {type(None)}", case_loggy.output)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),1)
		self.assertIsNone(current)


	def test_AddressField_no_validator(self):
		class CurrentAddress(AddressField):

			field_value = ""
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "AddressField_no_validator"
				init_level	= 10

		self.test_case = CurrentAddress()
		with self.assertLogs("AddressField_no_validator", 10) as case_loggy : current = self.test_case()
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIsNone(current)
		self.assertIn(

			f"INFO:AddressField_no_validator:{self.test_case} cannot validate email address",
			case_loggy.output
		)








	def test_SenderField_nomake(self):

		class pseudo_sender(SenderField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-sender-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-sender-top", 10) as case_loggy:

			self.test_case = pseudo_sender()
			current = self.test_case.make_field("sender")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-sender-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-sender-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-sender"
				init_level	= 10

			class pseudo_sender(SenderField):	pass

		with self.assertLogs("PseudoBuilder-sender", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_sender.make_field("sender")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-sender:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-sender:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "Builder-sender"
				init_level	= 10

			class pseudo_sender(SenderField):	pass

		with self.assertLogs("Builder-sender", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_sender.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder-sender:Putted \"sender\", \"{[ self.test_case.pseudo_sender ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder-sender:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)








	def test_RecipientField_nomake(self):

		class pseudo_recip(RecipientField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-recip-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-recip-top", 10) as case_loggy:

			self.test_case = pseudo_recip()
			current = self.test_case.make_field("recipient")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-recip-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-recip-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-recip"
				init_level	= 10

			class pseudo_recip(RecipientField):	pass

		with self.assertLogs("PseudoBuilder-recip", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_recip.make_field("recipient")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-recip:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-recip:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "Builder"
				init_level	= 10

			class pseudo_recip(RecipientField):	pass

		with self.assertLogs("Builder", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_recip.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder:Putted \"recipient\", \"{[ self.test_case.pseudo_recip ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)








	def test_ccField_nomake(self):

		class pseudo_сс(ccField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-cc-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-cc-top", 10) as case_loggy:

			self.test_case = pseudo_сс()
			current = self.test_case.make_field("cc")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-cc-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-cc-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-cc"
				init_level	= 10

			class pseudo_сс(ccField):	pass

		with self.assertLogs("PseudoBuilder-cc", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_сс.make_field("cc")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-cc:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-cc:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "Builder"
				init_level	= 10

			class pseudo_сс(ccField):	pass

		with self.assertLogs("Builder", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_сс.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder:Putted \"cc\", \"{[ self.test_case.pseudo_сс ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)








	def test_bccField_nomake(self):

		class pseudo_bсс(bccField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-bcc-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-bcc-top", 10) as case_loggy:

			self.test_case = pseudo_bсс()
			current = self.test_case.make_field("bcc")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-bcc-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-bcc-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-bcc"
				init_level	= 10

			class pseudo_bсс(bccField):	pass

		with self.assertLogs("PseudoBuilder-bcc", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_bсс.make_field("bcc")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-bcc:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-bcc:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "Builder"
				init_level	= 10

			class pseudo_bсс(bccField):	pass

		with self.assertLogs("Builder", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_bсс.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder:Putted \"bcc\", \"{[ self.test_case.pseudo_bсс ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)








	def test_TextField_simple_valid(self):

		for sample in (

			self.SAMPLE_TEXT_1,
			self.SAMPLE_TEXT_2,
			self.SAMPLE_TEXT_3,
			self.SAMPLE_TEXT_4,
			self.SAMPLE_TEXT_5,
			self.SAMPLE_TEXT_6,
			self.SAMPLE_TEXT_7,
			self.SAMPLE_TEXT_8,
		):

			slen = len(sample)
			self.current_text.field_value = sample

			with self.assertLogs("CurrentText", 10) as case_loggy : current = self.current_text()

			self.assertIn(f"DEBUG:CurrentText:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentText:Fetched {slen} symbol{flagrate(slen)}", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentText:Proper modifiers not found", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentText:Obtained {slen} symbol{flagrate(slen)}", case_loggy.output)
			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),4)
			self.assertEqual(current, sample)


	def test_TextField_format_valid(self):

		self.current_text.modifiers = self.SAMPLE_MODIFS

		for sample in (

			self.SAMPLE_TEXT_1,
			self.SAMPLE_TEXT_2,
			self.SAMPLE_TEXT_3,
			self.SAMPLE_TEXT_4,
			self.SAMPLE_TEXT_5,
			self.SAMPLE_TEXT_6,
			self.SAMPLE_TEXT_7,
			self.SAMPLE_TEXT_8,
		):
			slen = len(sample)
			fsample = sample.format(**self.SAMPLE_MODIFS)
			flen = len(fsample)
			self.current_text.field_value = sample

			with self.assertLogs("CurrentText", 10) as case_loggy : current = self.current_text()

			self.assertIn(f"DEBUG:CurrentText:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentText:Fetched {slen} symbol{flagrate(slen)}", case_loggy.output)
			self.assertIn(

				f"DEBUG:CurrentText:Obtained {flen} symbol{flagrate(slen)}",
				case_loggy.output
			)
			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertEqual(len(case_loggy.output),3)
			self.assertEqual(current, fsample)

		delattr(self.current_text, "modifiers")


	def test_TextField_format_KeyError(self):

		self.current_text.modifiers = { "first": "OOH" }

		for sample in ( self.SAMPLE_TEXT_7, self.SAMPLE_TEXT_8 ):

			slen = len(sample)
			self.current_text.field_value = sample

			with self.assertLogs("CurrentText", 10) as case_loggy : current = self.current_text()

			self.assertIsNone(current)
			self.assertEqual(len(case_loggy.output),3)
			self.no_loggy_levels(case_loggy.output, 50)
			self.assertIn(f"DEBUG:CurrentText:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentText:Fetched {slen} symbol{flagrate(slen)}", case_loggy.output)
			self.assertIn(

				f"ERROR:CurrentText:{self.current_text} field failed due to "
				"KeyError: 'second'",
				case_loggy.output
			)

		delattr(self.current_text, "modifiers")


	def test_TextField_invalid_field(self):

		self.current_text.modifiers = self.SAMPLE_MODIFS

		for invalid_value in ( 1, 1., True, False, print, LibraryContrib, ):
			with self.subTest(value=invalid_value):
				self.current_text.field_value = invalid_value

				with self.assertLogs("CurrentText", 10) as case_loggy : current = self.current_text()

				self.assertIsNone(current)
				self.assertIn(
					f"DEBUG:CurrentText:Fetched value of {type(invalid_value)}", case_loggy.output
				)
				self.assertIn(

					f"ERROR:CurrentText:Field {self.current_text} value must be string",
					case_loggy.output
				)

		self.current_text.field_value = None
		with self.assertLogs("CurrentText", 10) as case_loggy : current = self.current_text()
		self.assertIn(f"DEBUG:CurrentText:Fetched value of {type(None)}", case_loggy.output)
		self.assertIn("DEBUG:CurrentText:No value fetched for field", case_loggy.output)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),2)
		self.assertIsNone(current)
		delattr(self.current_text, "modifiers")


	def test_TextField_novalue(self):

		class CurrentText(TextField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "TextField_novalue"
				init_level	= 10

		self.test_case = CurrentText()
		with self.assertLogs("TextField_novalue", 10) as case_loggy : current = self.test_case()

		self.no_loggy_levels(case_loggy.output, 50)
		self.assertEqual(len(case_loggy.output),2)
		self.assertIsNone(current)
		self.assertIn("DEBUG:TextField_novalue:No value fetched for field", case_loggy.output)
		self.assertIn(
			f"ERROR:TextField_novalue:Field {self.current_text} must have value",case_loggy.output
		)


	def test_TextField_invalid_modifiers(self):

		self.current_text.field_value = "Bypass"

		for invalid_modifiers in (

			1, 1., True, False, print, LibraryContrib,
			({ "first": "OOH", "second": "WALLA" },),
			[{ "first": "OOH", "second": "WALLA" }],
			[ "first", "OOH", "second", "WALLA" ],
			( "first", "OOH", "second", "WALLA", ),
			{ "first", "OOH", "second", "WALLA" },
		):
			self.current_text.modifiers = invalid_modifiers

			with self.assertLogs("CurrentText", 10) as case_loggy : current = self.current_text()

			self.assertEqual(current, "Bypass")
			self.assertEqual(len(case_loggy.output),4)
			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertIn(f"DEBUG:CurrentText:Fetched value of {type('')}", case_loggy.output)
			self.assertIn("DEBUG:CurrentText:Fetched 6 symbols", case_loggy.output)
			self.assertIn("DEBUG:CurrentText:Proper modifiers not found", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentText:Obtained 6 symbols", case_loggy.output)

		delattr(self.current_text, "modifiers")








	def test_SubjectField_nomake(self):

		class pseudo_subj(SubjectField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-subject-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-subject-top", 10) as case_loggy:

			self.test_case = pseudo_subj()
			current = self.test_case.make_field("subject")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-subject-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-subject-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-subject"
				init_level	= 10

			class pseudo_subj(SubjectField):	pass

		with self.assertLogs("PseudoBuilder-subject", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_subj.make_field("subject")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-subject:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-subject:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "Builder"
				init_level	= 10

			class pseudo_subj(SubjectField):	pass

		with self.assertLogs("Builder", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_subj.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder:Putted \"subject\", \"{[ self.test_case.pseudo_subj ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)








	def test_BodyField_nomake(self):

		class pseudo_body(BodyField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-body-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-body-top", 10) as case_loggy:

			self.test_case = pseudo_body()
			current = self.test_case.make_field("body")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-body-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-body-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "PseudoBuilder-body"
				init_level	= 10

			class pseudo_body(BodyField):	pass

		with self.assertLogs("PseudoBuilder-body", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_body.make_field("body")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-body:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-body:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_HANDLER
				init_name	= "Builder"
				init_level	= 10

			class pseudo_body(BodyField):	pass

		with self.assertLogs("Builder", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_body.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder:Putted \"body\", \"{[ self.test_case.pseudo_body ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
















class AttachmentFieldsCases(HedwigTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.HEDWIG_MAIL_LETTER_AHANDLER):

				cls.current_attach.loggy.close()
				os.remove(cls.HEDWIG_MAIL_LETTER_AHANDLER)

		if	os.path.isfile(cls.attach_original): os.remove(cls.attach_original)
		if	os.path.isfile(cls.attach_modified): os.remove(cls.attach_modified)

	@classmethod
	def setUpClass(cls):

		class CurrentAttachment(AttachmentField):
			class loggy(LibraryContrib):

				handler		= cls.HEDWIG_MAIL_LETTER_AHANDLER
				init_name	= "CurrentAttachment"
				init_level	= 10
		cls.current_attach	= CurrentAttachment()
		cls.attach_original	= str(cls.HEDWIG_ROOT /"origin.attach")
		cls.attach_modified	= str(cls.HEDWIG_ROOT /"modified.attach")
		cls.attach_notexist	= str(cls.HEDWIG_ROOT /"notexist.attach")
		cls.make_loggy_file(cls, cls.HEDWIG_MAIL_LETTER_AHANDLER)

		cls.ors	= len(cls.attach_original)
		cls.mods= len(cls.attach_modified)
		cls.noe	= len(cls.attach_notexist)


	def valid_attachment_modifier(self, source :str) -> str : return self.attach_modified
	def setUp(self):

		if	not os.path.isfile(self.attach_original):	self.fmake(self.attach_original)
		if	not os.path.isfile(self.attach_modified):	self.fmake(self.attach_modified)
		if	hasattr(self.current_attach, "modifiers"):	delattr(self.current_attach, "modifiers")




	def test_AttachmentField_valid(self):

		self.current_attach.field_value = self.attach_original
		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()
		self.assertEqual(current, self.attach_original)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),4)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched {self.ors} symbols", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Modifiers not used", case_loggy.output)
		self.assertIn(
			f"DEBUG:CurrentAttachment:Obtained file path \"{self.attach_original}\"", case_loggy.output
		)


	def test_AttachmentField_modified_valid(self):

		self.current_attach.field_value	= self.attach_original
		self.current_attach.modifiers	= self.valid_attachment_modifier
		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),3)
		self.assertEqual(current, self.attach_modified)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched {self.ors} symbols", case_loggy.output)
		self.assertIn(
			f"DEBUG:CurrentAttachment:Obtained file path \"{self.attach_modified}\"", case_loggy.output
		)


	def test_AttachmentField_valid_notexist(self):

		self.current_attach.field_value = self.attach_notexist
		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()
		self.assertIsNone(current)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),4)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched {self.noe} symbols", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Modifiers not used", case_loggy.output)
		self.assertIn(

			f"INFO:CurrentAttachment:{self.current_attach} cannot attach "
			f"non-existent file \"{self.attach_notexist}\"",
			case_loggy.output
		)


	def test_AttachmentField_modified_valid_notexist_to_exist(self):

		self.current_attach.field_value	= self.attach_notexist
		self.current_attach.modifiers	= self.valid_attachment_modifier
		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),3)
		self.assertEqual(current, self.attach_modified)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched {self.noe} symbols", case_loggy.output)
		self.assertIn(
			f"DEBUG:CurrentAttachment:Obtained file path \"{self.attach_modified}\"", case_loggy.output
		)


	def test_AttachmentField_modified_valid_notexist(self):

		self.current_attach.field_value	= self.attach_notexist
		self.current_attach.modifiers	= lambda E : E
		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),3)
		self.assertIsNone(current)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched {self.noe} symbols", case_loggy.output)
		self.assertIn(

			f"INFO:CurrentAttachment:{self.current_attach} cannot attach "
			f"non-existent file \"{self.attach_notexist}\"",
			case_loggy.output
		)


	def test_AttachmentField_invalid_field(self):

		for invalid_value in ( 1, 1., True, False, print, LibraryContrib, ):
			with self.subTest(value=invalid_value):
				self.current_attach.field_value = invalid_value

				with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()

				self.no_loggy_levels(case_loggy.output, 50)
				self.assertEqual(len(case_loggy.output),2)
				self.assertIsNone(current)
				self.assertIn(
					f"DEBUG:CurrentAttachment:Fetched value of {type(invalid_value)}", case_loggy.output
				)
				self.assertIn(

					f"ERROR:CurrentAttachment:Field {self.current_attach} value must be string",
					case_loggy.output
				)

		self.current_attach.field_value = None
		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type(None)}", case_loggy.output)
		self.assertIn("DEBUG:CurrentAttachment:No value fetched for field", case_loggy.output)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertEqual(len(case_loggy.output),2)
		self.assertIsNone(current)


	def test_AttachmentField_novalue(self):

		class CurrentAttachment(AttachmentField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_AHANDLER
				init_name	= "CurrentAttachment_novalue"
				init_level	= 10

		self.test_case = CurrentAttachment()
		with self.assertLogs("CurrentAttachment_novalue", 10) as case_loggy : current = self.test_case()

		self.no_loggy_levels(case_loggy.output, 50)
		self.assertEqual(len(case_loggy.output),2)
		self.assertIsNone(current)
		self.assertIn("DEBUG:CurrentAttachment_novalue:No value fetched for field", case_loggy.output)
		self.assertIn(
			f"ERROR:CurrentAttachment_novalue:Field {self.test_case} must have value",case_loggy.output
		)


	def test_AttachmentField_invalid_modifiers(self):

		self.current_attach.field_value = "Bypass"

		for invalid_modifiers in (

			1, 1., True, False,
			( self.valid_attachment_modifier,),
			[ self.valid_attachment_modifier ],
			{ self.valid_attachment_modifier },
			{ "modifiers": self.valid_attachment_modifier },
		):
			self.current_attach.modifiers = invalid_modifiers

			with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()

			self.assertIsNone(current)
			self.assertEqual(len(case_loggy.output),4)
			self.no_loggy_levels(case_loggy.output, 30,40,50)
			self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentAttachment:Fetched 6 symbols", case_loggy.output)
			self.assertIn(f"DEBUG:CurrentAttachment:Modifiers not used", case_loggy.output)
			self.assertIn(

				f"INFO:CurrentAttachment:{self.current_attach} cannot attach "
				f"non-existent file \"Bypass\"",
				case_loggy.output
			)


	def test_AttachmentField_modifiers_error(self):

		self.current_attach.field_value = "Bypass"
		self.current_attach.modifiers = lambda : "Bypass"

		with self.assertLogs("CurrentAttachment", 10) as case_loggy : current = self.current_attach()

		self.assertIsNone(current)
		self.assertEqual(len(case_loggy.output),4)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:CurrentAttachment:Fetched 6 symbols", case_loggy.output)
		self.assertIn(

			f"ERROR:CurrentAttachment:AttachmentField {self.current_attach} field failed due to "
			f"TypeError: {self.current_attach.modifiers.__qualname__}() "
			"takes 0 positional arguments but 1 was given",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:CurrentAttachment:{self.current_attach} cannot attach "
			f"non-existent file \"Bypass\"",
			case_loggy.output
		)




	def test_AttachmentField_nomake(self):

		class pseudo_attach(AttachmentField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_AHANDLER
				init_name	= "PseudoBuilder-attach-top"
				init_level	= 10


		with self.assertLogs("PseudoBuilder-attach-top", 10) as case_loggy:

			self.test_case = pseudo_attach()
			current = self.test_case.make_field("attach")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-attach-top:Builder not found",case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:PseudoBuilder-attach-top:Builder not found"),2)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class PseudoBuilder(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_AHANDLER
				init_name	= "PseudoBuilder-attach"
				init_level	= 10

			class pseudo_attach(AttachmentField):	pass

		with self.assertLogs("PseudoBuilder-attach", 10) as case_loggy:

			self.test_case = PseudoBuilder()
			current = self.test_case.pseudo_attach.make_field("attach")

		self.assertIsNone(current)
		self.assertIn(f"DEBUG:PseudoBuilder-attach:Improper builder {self.test_case}",case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:PseudoBuilder-attach:Improper builder {self.test_case}"),2
		)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.test_case.loggy.close()


		class Builder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_LETTER_AHANDLER
				init_name	= "Builder"
				init_level	= 10

			class pseudo_attach(AttachmentField):	pass

		with self.assertLogs("Builder", 10) as case_loggy:

			self.test_case = Builder()
			current = self.test_case.pseudo_attach.make_field(42)

		self.assertIsNone(current)
		self.assertIn(

			f"DEBUG:Builder:Putted \"attachment\", \"{[ self.test_case.pseudo_attach ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:Builder:Field name \"42\" must be a string",case_loggy.output)
		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pygwarts.tests.hagrid					import EasySet
from	pygwarts.irma.contrib					import LibraryContrib
from	pygwarts.irma.shelve					import LibraryShelf
from	pygwarts.hagrid.thrivables				import Copse
from	pygwarts.hagrid.sprouts					import fssprout
from	pygwarts.hagrid.planting				import Flourish
from	pygwarts.hagrid.cultivation.registering	import PlantRegister
from	pygwarts.hagrid.cultivation.registering	import PlantRegisterQuerier
from	pygwarts.hagrid.cultivation.registering	import WG
from	pygwarts.hagrid.cultivation.registering	import TG
from	pygwarts.hagrid.cultivation.registering	import LG
from	pygwarts.hagrid.cultivation.registering	import WL
from	pygwarts.hagrid.cultivation.registering	import TL
from	pygwarts.hagrid.cultivation.registering	import LL








class RegisterSingleCase(EasySet):

	"""
		Single source registering and querying
	"""

	@classmethod
	def tearDownClass(cls): cls.clean(cls)
	@classmethod
	def setUpClass(cls):

		super().setUpClass()

		class Forest(Copse):

			@fssprout(cls.EASY_SET_SPROUT)
			@PlantRegister("garden")
			class walk(Flourish):		pass
			class garden(LibraryShelf):	pass

		cls.test_case = Forest()
		cls.test_case.walk()
		cls.prq = PlantRegisterQuerier(cls.test_case.garden)
		cls.miss_branch = os.path.join(cls.EASY_SET_SPROUT, "miss branch")
		cls.test_case.garden.produce(cls.EASY_SET_SEEDS, ignore_mod=True)
		cls.extern_shelf = LibraryShelf()
		cls.extern_shelf.grab(cls.EASY_SET_SEEDS)
		cls.prq2 = PlantRegisterQuerier(cls.extern_shelf)


	def test_PRQ_bad_init(self):

		for garden in (

			1, 1., "LibraryShelf", LibraryShelf, True, False, None, ..., print,
			[ self.test_case.garden ],
			( self.test_case.garden, ),
			{ self.test_case.garden },
			{ "garden": self.test_case.garden },
		):
			self.assertIsNone(PlantRegisterQuerier(garden).garden)


	def test_PRQ_query_bad_branch(self):

		for branch in (

			1, 1., "folder", LibraryShelf, True, False, ..., print,
			[ self.EASY_SET_SPROUT ],
			( self.EASY_SET_SPROUT, ),
			{ self.EASY_SET_SPROUT },
			{ "branch": self.EASY_SET_SPROUT },
		):
			with self.subTest(branch=branch):

				self.assertIsNone(self.prq.query(branch, key=WG))
				self.assertIsNone(self.prq.query(branch, key=WL))
				self.assertIsNone(self.prq.query(branch, key=TG))
				self.assertIsNone(self.prq.query(branch, key=LG))
				self.assertIsNone(self.prq.query(branch, key=TL))
				self.assertIsNone(self.prq.query(branch, key=LL))


	def test_PRQ_query_bad_sprout(self):

		for sprout in (

			1, 1., "folder", LibraryShelf, True, False, ..., print,
			[ self.EASY_SET_SPROUT ],
			( self.EASY_SET_SPROUT, ),
			{ self.EASY_SET_SPROUT },
			{ "sprout": self.EASY_SET_SPROUT },
		):
			with self.subTest(sprout=sprout):

				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, sprout, WG))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, sprout, WL))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, sprout, TG))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, sprout, LG))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, sprout, TL))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, sprout, LL))


	def test_PRQ_query_bad_key(self):

		for key in (

			1, 1., "weight", LibraryShelf, True, False, ..., print,
			[ WG ],( WG, ),{ WG },{ "key": WG },
		):
			with self.subTest(key=key):

				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key))
				self.assertIsNone(self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key))


	def test_PRQ_bad_named_queries(self):

		for shoot in (

			1, 1., "folder", LibraryShelf, True, False, ..., print,
			[ self.EASY_SET_SPROUT ],
			( self.EASY_SET_SPROUT, ),
			{ self.EASY_SET_SPROUT },
			{ "shoot": self.EASY_SET_SPROUT },
		):
			with self.subTest(shoot=shoot):

				self.assertIsNone(self.prq.WG(shoot))
				self.assertIsNone(self.prq.WG(shoot, shoot))
				self.assertIsNone(self.prq.WG(sprout=shoot))

				self.assertIsNone(self.prq.TG(shoot))
				self.assertIsNone(self.prq.TG(shoot, shoot))
				self.assertIsNone(self.prq.TG(sprout=shoot))

				self.assertIsNone(self.prq.LG(shoot))
				self.assertIsNone(self.prq.LG(shoot, shoot))
				self.assertIsNone(self.prq.LG(sprout=shoot))

				self.assertIsNone(self.prq.WL(shoot))
				self.assertIsNone(self.prq.WL(shoot, shoot))
				self.assertIsNone(self.prq.WL(sprout=shoot))

				self.assertIsNone(self.prq.TL(shoot))
				self.assertIsNone(self.prq.TL(shoot, shoot))
				self.assertIsNone(self.prq.TL(sprout=shoot))

				self.assertIsNone(self.prq.LL(shoot))
				self.assertIsNone(self.prq.LL(shoot, shoot))
				self.assertIsNone(self.prq.LL(sprout=shoot))








	def test_PRQ_bad_garden(self):

		for garden in (

			1, 1., "LibraryShelf", LibraryShelf, True, False, None, ..., print,
			[ self.test_case.garden ],
			( self.test_case.garden, ),
			{ self.test_case.garden },
			{ "garden": self.test_case.garden },
		):
			prq = PlantRegisterQuerier(garden)

			self.assertIsNone(prq.query(key=WG))
			self.assertIsNone(prq.query(key=TG))
			self.assertIsNone(prq.query(key=LG))
			self.assertIsNone(prq.query(key=WL))
			self.assertIsNone(prq.query(key=TL))
			self.assertIsNone(prq.query(key=LL))

			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, key=WG))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, key=TG))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, key=LG))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, key=WL))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, key=TL))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, key=LL))

			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=WG))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=TG))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=LG))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=WL))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=TL))
			self.assertIsNone(prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=LL))

			self.assertIsNone(prq.query(sprout=self.EASY_SET_SPROUT, key=WG))
			self.assertIsNone(prq.query(sprout=self.EASY_SET_SPROUT, key=TG))
			self.assertIsNone(prq.query(sprout=self.EASY_SET_SPROUT, key=LG))
			self.assertIsNone(prq.query(sprout=self.EASY_SET_SPROUT, key=WL))
			self.assertIsNone(prq.query(sprout=self.EASY_SET_SPROUT, key=TL))
			self.assertIsNone(prq.query(sprout=self.EASY_SET_SPROUT, key=LL))

			self.assertIsNone(prq.WG())
			self.assertIsNone(prq.TG())
			self.assertIsNone(prq.LG())
			self.assertIsNone(prq.WL())
			self.assertIsNone(prq.TL())
			self.assertIsNone(prq.LL())
			self.assertIsNone(prq.WG(apparent=True))
			self.assertIsNone(prq.WL(apparent=True))

			self.assertIsNone(prq.WG(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.TG(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.LG(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.WL(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.TL(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.LL(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.WG(self.EASY_SET_SPROUT, apparent=True))
			self.assertIsNone(prq.WL(self.EASY_SET_SPROUT, apparent=True))

			self.assertIsNone(prq.WG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.TG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.LG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.WL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.TL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.LL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.WG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True))
			self.assertIsNone(prq.WL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True))

			self.assertIsNone(prq.WG(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.TG(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.LG(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.WL(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.TL(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.LL(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.WG(sprout=self.EASY_SET_SPROUT, apparent=True))
			self.assertIsNone(prq.WL(sprout=self.EASY_SET_SPROUT, apparent=True))

			self.assertIsNone(prq.content(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT))
			self.assertIsNone(prq.content(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True))
			self.assertIsNone(prq.content(self.EASY_SET_SPROUT))
			self.assertIsNone(prq.content(self.EASY_SET_SPROUT, apparent=True))
			self.assertIsNone(prq.content(sprout=self.EASY_SET_SPROUT))
			self.assertIsNone(prq.content(sprout=self.EASY_SET_SPROUT, apparent=True))
			self.assertIsNone(prq.content())
			self.assertIsNone(prq.content(apparent=True))








	def test_miss_branch(self):

		self.assertIsNone(self.prq.query(self.miss_branch, WG))
		self.assertIsNone(self.prq.query(self.miss_branch, TG))
		self.assertIsNone(self.prq.query(self.miss_branch, LG))
		self.assertIsNone(self.prq.query(self.miss_branch, WL))
		self.assertIsNone(self.prq.query(self.miss_branch, TL))
		self.assertIsNone(self.prq.query(self.miss_branch, LL))

		self.assertIsNone(self.prq.query(self.miss_branch, self.EASY_SET_SPROUT, WG))
		self.assertIsNone(self.prq.query(self.miss_branch, self.EASY_SET_SPROUT, TG))
		self.assertIsNone(self.prq.query(self.miss_branch, self.EASY_SET_SPROUT, LG))
		self.assertIsNone(self.prq.query(self.miss_branch, self.EASY_SET_SPROUT, WL))
		self.assertIsNone(self.prq.query(self.miss_branch, self.EASY_SET_SPROUT, TL))
		self.assertIsNone(self.prq.query(self.miss_branch, self.EASY_SET_SPROUT, LL))

		self.assertIsNone(self.prq.WG(self.miss_branch))
		self.assertIsNone(self.prq.TG(self.miss_branch))
		self.assertIsNone(self.prq.LG(self.miss_branch))
		self.assertIsNone(self.prq.WL(self.miss_branch))
		self.assertIsNone(self.prq.TL(self.miss_branch))
		self.assertIsNone(self.prq.LL(self.miss_branch))

		self.assertIsNone(self.prq.WG(self.miss_branch, self.EASY_SET_SPROUT))
		self.assertIsNone(self.prq.TG(self.miss_branch, self.EASY_SET_SPROUT))
		self.assertIsNone(self.prq.LG(self.miss_branch, self.EASY_SET_SPROUT))
		self.assertIsNone(self.prq.WL(self.miss_branch, self.EASY_SET_SPROUT))
		self.assertIsNone(self.prq.TL(self.miss_branch, self.EASY_SET_SPROUT))
		self.assertIsNone(self.prq.LL(self.miss_branch, self.EASY_SET_SPROUT))

		self.assertIsNone(self.prq.WG(self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.WL(self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.WG(self.miss_branch, self.EASY_SET_SPROUT, apparent=True))
		self.assertIsNone(self.prq.WL(self.miss_branch, self.EASY_SET_SPROUT, apparent=True))


	def test_miss_sprout(self):

		self.assertIsNone(self.prq.query(sprout=self.miss_branch, key=WG))
		self.assertIsNone(self.prq.query(sprout=self.miss_branch, key=TG))
		self.assertIsNone(self.prq.query(sprout=self.miss_branch, key=LG))
		self.assertIsNone(self.prq.query(sprout=self.miss_branch, key=WL))
		self.assertIsNone(self.prq.query(sprout=self.miss_branch, key=TL))
		self.assertIsNone(self.prq.query(sprout=self.miss_branch, key=LL))

		self.assertIsNone(self.prq.query(self.pros_folder, self.miss_branch, key=WG))
		self.assertIsNone(self.prq.query(self.pros_folder, self.miss_branch, key=TG))
		self.assertIsNone(self.prq.query(self.pros_folder, self.miss_branch, key=LG))
		self.assertIsNone(self.prq.query(self.pros_folder, self.miss_branch, key=WL))
		self.assertIsNone(self.prq.query(self.pros_folder, self.miss_branch, key=TL))
		self.assertIsNone(self.prq.query(self.pros_folder, self.miss_branch, key=LL))

		self.assertIsNone(self.prq.WG(sprout=self.miss_branch))
		self.assertIsNone(self.prq.TG(sprout=self.miss_branch))
		self.assertIsNone(self.prq.LG(sprout=self.miss_branch))
		self.assertIsNone(self.prq.WL(sprout=self.miss_branch))
		self.assertIsNone(self.prq.TL(sprout=self.miss_branch))
		self.assertIsNone(self.prq.LL(sprout=self.miss_branch))

		self.assertIsNone(self.prq.WG(self.pros_folder, self.miss_branch))
		self.assertIsNone(self.prq.TG(self.pros_folder, self.miss_branch))
		self.assertIsNone(self.prq.LG(self.pros_folder, self.miss_branch))
		self.assertIsNone(self.prq.WL(self.pros_folder, self.miss_branch))
		self.assertIsNone(self.prq.TL(self.pros_folder, self.miss_branch))
		self.assertIsNone(self.prq.LL(self.pros_folder, self.miss_branch))

		self.assertIsNone(self.prq.WG(sprout=self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.WL(sprout=self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.WG(self.pros_folder, self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.WL(self.pros_folder, self.miss_branch, apparent=True))








	def test_register_root_query(self):

		root_WG = self.prq.query(key=WG)
		root_TG = self.prq.query(key=TG)
		root_LG = self.prq.query(key=LG)
		root_WL = self.prq.query(key=WL)
		root_TL = self.prq.query(key=TL)
		root_LL = self.prq.query(key=LL)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		root2_WG = self.prq2.query(key=WG)
		root2_TG = self.prq2.query(key=TG)
		root2_LG = self.prq2.query(key=LG)
		root2_WL = self.prq2.query(key=WL)
		root2_TL = self.prq2.query(key=TL)
		root2_LL = self.prq2.query(key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)


	def test_branch_register_root_query(self):

		root_WG = self.prq.query(self.EASY_SET_SPROUT, key=WG)
		root_TG = self.prq.query(self.EASY_SET_SPROUT, key=TG)
		root_LG = self.prq.query(self.EASY_SET_SPROUT, key=LG)
		root_WL = self.prq.query(self.EASY_SET_SPROUT, key=WL)
		root_TL = self.prq.query(self.EASY_SET_SPROUT, key=TL)
		root_LL = self.prq.query(self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		root2_WG = self.prq2.query(self.EASY_SET_SPROUT, key=WG)
		root2_TG = self.prq2.query(self.EASY_SET_SPROUT, key=TG)
		root2_LG = self.prq2.query(self.EASY_SET_SPROUT, key=LG)
		root2_WL = self.prq2.query(self.EASY_SET_SPROUT, key=WL)
		root2_TL = self.prq2.query(self.EASY_SET_SPROUT, key=TL)
		root2_LL = self.prq2.query(self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)


	def test_branch_sprout_register_root_query(self):

		root_WG = self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=WG)
		root_TG = self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=TG)
		root_LG = self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=LG)
		root_WL = self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=WL)
		root_TL = self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=TL)
		root_LL = self.prq.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		root2_WG = self.prq2.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=WG)
		root2_TG = self.prq2.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=TG)
		root2_LG = self.prq2.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=LG)
		root2_WL = self.prq2.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=WL)
		root2_TL = self.prq2.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=TL)
		root2_LL = self.prq2.query(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)


	def test_sprout_register_root_query(self):

		root_WG = self.prq.query(sprout=self.EASY_SET_SPROUT, key=WG)
		root_TG = self.prq.query(sprout=self.EASY_SET_SPROUT, key=TG)
		root_LG = self.prq.query(sprout=self.EASY_SET_SPROUT, key=LG)
		root_WL = self.prq.query(sprout=self.EASY_SET_SPROUT, key=WL)
		root_TL = self.prq.query(sprout=self.EASY_SET_SPROUT, key=TL)
		root_LL = self.prq.query(sprout=self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		root2_WG = self.prq2.query(sprout=self.EASY_SET_SPROUT, key=WG)
		root2_TG = self.prq2.query(sprout=self.EASY_SET_SPROUT, key=TG)
		root2_LG = self.prq2.query(sprout=self.EASY_SET_SPROUT, key=LG)
		root2_WL = self.prq2.query(sprout=self.EASY_SET_SPROUT, key=WL)
		root2_TL = self.prq2.query(sprout=self.EASY_SET_SPROUT, key=TL)
		root2_LL = self.prq2.query(sprout=self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)








	def test_branch_register_child_query(self):

		child_WG = self.prq.query(self.pros_folder, key=WG)
		child_TG = self.prq.query(self.pros_folder, key=TG)
		child_LG = self.prq.query(self.pros_folder, key=LG)
		child_WL = self.prq.query(self.pros_folder, key=WL)
		child_TL = self.prq.query(self.pros_folder, key=TL)
		child_LL = self.prq.query(self.pros_folder, key=LL)

		self.assertIsInstance(child_WG, int)
		self.assertIsInstance(child_WL, int)
		self.assertEqual(child_TG, 1)
		self.assertEqual(child_LG, 4)
		self.assertEqual(child_TL, 1)
		self.assertEqual(child_LL, 3)

		child2_WG = self.prq2.query(self.pros_folder, key=WG)
		child2_TG = self.prq2.query(self.pros_folder, key=TG)
		child2_LG = self.prq2.query(self.pros_folder, key=LG)
		child2_WL = self.prq2.query(self.pros_folder, key=WL)
		child2_TL = self.prq2.query(self.pros_folder, key=TL)
		child2_LL = self.prq2.query(self.pros_folder, key=LL)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 1)
		self.assertEqual(child2_LG, 4)
		self.assertEqual(child2_TL, 1)
		self.assertEqual(child2_LL, 3)


	def test_branch_sprout_register_child_query(self):

		child_WG = self.prq.query(self.pros_folder, self.EASY_SET_SPROUT, key=WG)
		child_TG = self.prq.query(self.pros_folder, self.EASY_SET_SPROUT, key=TG)
		child_LG = self.prq.query(self.pros_folder, self.EASY_SET_SPROUT, key=LG)
		child_WL = self.prq.query(self.pros_folder, self.EASY_SET_SPROUT, key=WL)
		child_TL = self.prq.query(self.pros_folder, self.EASY_SET_SPROUT, key=TL)
		child_LL = self.prq.query(self.pros_folder, self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(child_WG, int)
		self.assertIsInstance(child_WL, int)
		self.assertEqual(child_TG, 1)
		self.assertEqual(child_LG, 4)
		self.assertEqual(child_TL, 1)
		self.assertEqual(child_LL, 3)

		child2_WG = self.prq2.query(self.pros_folder, self.EASY_SET_SPROUT, key=WG)
		child2_TG = self.prq2.query(self.pros_folder, self.EASY_SET_SPROUT, key=TG)
		child2_LG = self.prq2.query(self.pros_folder, self.EASY_SET_SPROUT, key=LG)
		child2_WL = self.prq2.query(self.pros_folder, self.EASY_SET_SPROUT, key=WL)
		child2_TL = self.prq2.query(self.pros_folder, self.EASY_SET_SPROUT, key=TL)
		child2_LL = self.prq2.query(self.pros_folder, self.EASY_SET_SPROUT, key=LL)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 1)
		self.assertEqual(child2_LG, 4)
		self.assertEqual(child2_TL, 1)
		self.assertEqual(child2_LL, 3)








	def test_register_root_stats(self):

		root_WG = self.prq.WG()
		root_TG = self.prq.TG()
		root_LG = self.prq.LG()
		root_WL = self.prq.WL()
		root_TL = self.prq.TL()
		root_LL = self.prq.LL()

		root_WG_apparent = self.prq.WG(apparent=True)
		root_WL_apparent = self.prq.WL(apparent=True)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		self.assertIsInstance(root_WG_apparent, int)
		self.assertIsInstance(root_WL_apparent, int)

		self.assertEqual(root_WG_apparent - root_WG, 4 *4096)
		self.assertEqual(root_WL_apparent - root_WL, 3 *4096)

		root2_WG = self.prq2.WG()
		root2_TG = self.prq2.TG()
		root2_LG = self.prq2.LG()
		root2_WL = self.prq2.WL()
		root2_TL = self.prq2.TL()
		root2_LL = self.prq2.LL()

		root2_WG_apparent = self.prq2.WG(apparent=True)
		root2_WL_apparent = self.prq2.WL(apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4 *4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 3 *4096)


	def test_branch_register_root_stats(self):

		root_WG = self.prq.WG(self.EASY_SET_SPROUT)
		root_TG = self.prq.TG(self.EASY_SET_SPROUT)
		root_LG = self.prq.LG(self.EASY_SET_SPROUT)
		root_WL = self.prq.WL(self.EASY_SET_SPROUT)
		root_TL = self.prq.TL(self.EASY_SET_SPROUT)
		root_LL = self.prq.LL(self.EASY_SET_SPROUT)

		root_WG_apparent = self.prq.WG(self.EASY_SET_SPROUT, apparent=True)
		root_WL_apparent = self.prq.WL(self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		self.assertIsInstance(root_WG_apparent, int)
		self.assertIsInstance(root_WL_apparent, int)

		self.assertEqual(root_WG_apparent - root_WG, 4 *4096)
		self.assertEqual(root_WL_apparent - root_WL, 3 *4096)

		root2_WG = self.prq2.WG(self.EASY_SET_SPROUT)
		root2_TG = self.prq2.TG(self.EASY_SET_SPROUT)
		root2_LG = self.prq2.LG(self.EASY_SET_SPROUT)
		root2_WL = self.prq2.WL(self.EASY_SET_SPROUT)
		root2_TL = self.prq2.TL(self.EASY_SET_SPROUT)
		root2_LL = self.prq2.LL(self.EASY_SET_SPROUT)

		root2_WG_apparent = self.prq2.WG(self.EASY_SET_SPROUT, apparent=True)
		root2_WL_apparent = self.prq2.WL(self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4 *4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 3 *4096)


	def test_branch_sprout_register_root_stats(self):

		root_WG = self.prq.WG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root_TG = self.prq.TG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root_LG = self.prq.LG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root_WL = self.prq.WL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root_TL = self.prq.TL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root_LL = self.prq.LL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)

		root_WG_apparent = self.prq.WG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True)
		root_WL_apparent = self.prq.WL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		self.assertIsInstance(root_WG_apparent, int)
		self.assertIsInstance(root_WL_apparent, int)

		self.assertEqual(root_WG_apparent - root_WG, 4 *4096)
		self.assertEqual(root_WL_apparent - root_WL, 3 *4096)

		root2_WG = self.prq2.WG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root2_TG = self.prq2.TG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root2_LG = self.prq2.LG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root2_WL = self.prq2.WL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root2_TL = self.prq2.TL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)
		root2_LL = self.prq2.LL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT)

		root2_WG_apparent = self.prq2.WG(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True)
		root2_WL_apparent = self.prq2.WL(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4 *4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 3 *4096)


	def test_sprout_register_root_stats(self):

		root_WG = self.prq.WG(sprout=self.EASY_SET_SPROUT)
		root_TG = self.prq.TG(sprout=self.EASY_SET_SPROUT)
		root_LG = self.prq.LG(sprout=self.EASY_SET_SPROUT)
		root_WL = self.prq.WL(sprout=self.EASY_SET_SPROUT)
		root_TL = self.prq.TL(sprout=self.EASY_SET_SPROUT)
		root_LL = self.prq.LL(sprout=self.EASY_SET_SPROUT)

		root_WG_apparent = self.prq.WG(sprout=self.EASY_SET_SPROUT, apparent=True)
		root_WL_apparent = self.prq.WL(sprout=self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(root_WG, int)
		self.assertIsInstance(root_WL, int)
		self.assertEqual(root_TG, 4)
		self.assertEqual(root_LG, 8)
		self.assertEqual(root_TL, 3)
		self.assertEqual(root_LL, 1)

		self.assertIsInstance(root_WG_apparent, int)
		self.assertIsInstance(root_WL_apparent, int)

		self.assertEqual(root_WG_apparent - root_WG, 4 *4096)
		self.assertEqual(root_WL_apparent - root_WL, 3 *4096)

		root2_WG = self.prq2.WG(sprout=self.EASY_SET_SPROUT)
		root2_TG = self.prq2.TG(sprout=self.EASY_SET_SPROUT)
		root2_LG = self.prq2.LG(sprout=self.EASY_SET_SPROUT)
		root2_WL = self.prq2.WL(sprout=self.EASY_SET_SPROUT)
		root2_TL = self.prq2.TL(sprout=self.EASY_SET_SPROUT)
		root2_LL = self.prq2.LL(sprout=self.EASY_SET_SPROUT)

		root2_WG_apparent = self.prq2.WG(sprout=self.EASY_SET_SPROUT, apparent=True)
		root2_WL_apparent = self.prq2.WL(sprout=self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 4)
		self.assertEqual(root2_LG, 8)
		self.assertEqual(root2_TL, 3)
		self.assertEqual(root2_LL, 1)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4 *4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 3 *4096)








	def test_branch_register_child_stats(self):

		child_WG = self.prq.WG(self.pros_folder)
		child_TG = self.prq.TG(self.pros_folder)
		child_LG = self.prq.LG(self.pros_folder)
		child_WL = self.prq.WL(self.pros_folder)
		child_TL = self.prq.TL(self.pros_folder)
		child_LL = self.prq.LL(self.pros_folder)

		child_WG_apparent = self.prq.WG(self.pros_folder, apparent=True)
		child_WL_apparent = self.prq.WL(self.pros_folder, apparent=True)

		self.assertIsInstance(child_WG, int)
		self.assertIsInstance(child_WL, int)
		self.assertEqual(child_TG, 1)
		self.assertEqual(child_LG, 4)
		self.assertEqual(child_TL, 1)
		self.assertEqual(child_LL, 3)

		self.assertIsInstance(child_WG_apparent, int)
		self.assertIsInstance(child_WL_apparent, int)

		self.assertEqual(child_WG_apparent - child_WG, 4096)
		self.assertEqual(child_WL_apparent - child_WL, 4096)

		child2_WG = self.prq2.WG(self.pros_folder)
		child2_TG = self.prq2.TG(self.pros_folder)
		child2_LG = self.prq2.LG(self.pros_folder)
		child2_WL = self.prq2.WL(self.pros_folder)
		child2_TL = self.prq2.TL(self.pros_folder)
		child2_LL = self.prq2.LL(self.pros_folder)

		child2_WG_apparent = self.prq2.WG(self.pros_folder, apparent=True)
		child2_WL_apparent = self.prq2.WL(self.pros_folder, apparent=True)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 1)
		self.assertEqual(child2_LG, 4)
		self.assertEqual(child2_TL, 1)
		self.assertEqual(child2_LL, 3)

		self.assertIsInstance(child2_WG_apparent, int)
		self.assertIsInstance(child2_WL_apparent, int)

		self.assertEqual(child2_WG_apparent - child2_WG, 4096)
		self.assertEqual(child2_WL_apparent - child2_WL, 4096)


	def test_branch_sprout_register_child_stats(self):

		child_WG = self.prq.WG(self.pros_folder, self.EASY_SET_SPROUT)
		child_TG = self.prq.TG(self.pros_folder, self.EASY_SET_SPROUT)
		child_LG = self.prq.LG(self.pros_folder, self.EASY_SET_SPROUT)
		child_WL = self.prq.WL(self.pros_folder, self.EASY_SET_SPROUT)
		child_TL = self.prq.TL(self.pros_folder, self.EASY_SET_SPROUT)
		child_LL = self.prq.LL(self.pros_folder, self.EASY_SET_SPROUT)

		child_WG_apparent = self.prq.WG(self.pros_folder, self.EASY_SET_SPROUT, apparent=True)
		child_WL_apparent = self.prq.WL(self.pros_folder, self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(child_WG, int)
		self.assertIsInstance(child_WL, int)
		self.assertEqual(child_TG, 1)
		self.assertEqual(child_LG, 4)
		self.assertEqual(child_TL, 1)
		self.assertEqual(child_LL, 3)

		self.assertIsInstance(child_WG_apparent, int)
		self.assertIsInstance(child_WL_apparent, int)

		self.assertEqual(child_WG_apparent - child_WG, 4096)
		self.assertEqual(child_WL_apparent - child_WL, 4096)

		child2_WG = self.prq2.WG(self.pros_folder, self.EASY_SET_SPROUT)
		child2_TG = self.prq2.TG(self.pros_folder, self.EASY_SET_SPROUT)
		child2_LG = self.prq2.LG(self.pros_folder, self.EASY_SET_SPROUT)
		child2_WL = self.prq2.WL(self.pros_folder, self.EASY_SET_SPROUT)
		child2_TL = self.prq2.TL(self.pros_folder, self.EASY_SET_SPROUT)
		child2_LL = self.prq2.LL(self.pros_folder, self.EASY_SET_SPROUT)

		child2_WG_apparent = self.prq2.WG(self.pros_folder, self.EASY_SET_SPROUT, apparent=True)
		child2_WL_apparent = self.prq2.WL(self.pros_folder, self.EASY_SET_SPROUT, apparent=True)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 1)
		self.assertEqual(child2_LG, 4)
		self.assertEqual(child2_TL, 1)
		self.assertEqual(child2_LL, 3)

		self.assertIsInstance(child2_WG_apparent, int)
		self.assertIsInstance(child2_WL_apparent, int)

		self.assertEqual(child2_WG_apparent - child2_WG, 4096)
		self.assertEqual(child2_WL_apparent - child2_WL, 4096)








	def test_branch_sprout_content_root(self):

		for content in (

			self.prq.content(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT),
			self.prq.content(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True),
			self.prq2.content(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT),
			self.prq2.content(self.EASY_SET_SPROUT, self.EASY_SET_SPROUT, apparent=True),
		):
			self.assertEqual(len(content),4)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertIsInstance(content[2],tuple)
			self.assertIsInstance(content[3],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertEqual(len(content[2]),2)
			self.assertEqual(len(content[3]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)
			self.assertIsInstance(content[2][0],int)
			self.assertIsInstance(content[2][1],str)
			self.assertIsInstance(content[3][0],int)
			self.assertIsInstance(content[3][1],str)


	def test_branch_content_root(self):

		for content in (

			self.prq.content(self.EASY_SET_SPROUT),
			self.prq.content(self.EASY_SET_SPROUT, apparent=True),
			self.prq2.content(self.EASY_SET_SPROUT),
			self.prq2.content(self.EASY_SET_SPROUT, apparent=True),
		):
			self.assertEqual(len(content),4)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertIsInstance(content[2],tuple)
			self.assertIsInstance(content[3],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertEqual(len(content[2]),2)
			self.assertEqual(len(content[3]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)
			self.assertIsInstance(content[2][0],int)
			self.assertIsInstance(content[2][1],str)
			self.assertIsInstance(content[3][0],int)
			self.assertIsInstance(content[3][1],str)


	def test_sprout_content_root(self):

		for content in (

			self.prq.content(sprout=self.EASY_SET_SPROUT),
			self.prq.content(sprout=self.EASY_SET_SPROUT, apparent=True),
			self.prq2.content(sprout=self.EASY_SET_SPROUT),
			self.prq2.content(sprout=self.EASY_SET_SPROUT, apparent=True),
		):
			self.assertEqual(len(content),4)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertIsInstance(content[2],tuple)
			self.assertIsInstance(content[3],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertEqual(len(content[2]),2)
			self.assertEqual(len(content[3]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)
			self.assertIsInstance(content[2][0],int)
			self.assertIsInstance(content[2][1],str)
			self.assertIsInstance(content[3][0],int)
			self.assertIsInstance(content[3][1],str)


	def test_content_root(self):

		for content in (

			self.prq.content(),
			self.prq.content(apparent=True),
			self.prq2.content(),
			self.prq2.content(apparent=True),
		):
			self.assertEqual(len(content),4)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertIsInstance(content[2],tuple)
			self.assertIsInstance(content[3],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertEqual(len(content[2]),2)
			self.assertEqual(len(content[3]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)
			self.assertIsInstance(content[2][0],int)
			self.assertIsInstance(content[2][1],str)
			self.assertIsInstance(content[3][0],int)
			self.assertIsInstance(content[3][1],str)


	def test_content_bad_shoots(self):

		for shoot in (

			1, 1., "folder", LibraryShelf, True, False, ..., print,
			[ self.EASY_SET_SPROUT ],
			( self.EASY_SET_SPROUT, ),
			{ self.EASY_SET_SPROUT },
			{ "shoot": self.EASY_SET_SPROUT },
		):
			with self.subTest(shoot=shoot):

				self.assertIsNone(self.prq.content(shoot))
				self.assertIsNone(self.prq.content(shoot, shoot))
				self.assertIsNone(self.prq.content(sprout=shoot))
				self.assertIsNone(self.prq.content(shoot, apparent=True))
				self.assertIsNone(self.prq.content(shoot, shoot, apparent=True))
				self.assertIsNone(self.prq.content(sprout=shoot, apparent=True))


	def test_content_miss_branch(self):

		self.assertIsNone(self.prq.content(self.miss_branch))
		self.assertIsNone(self.prq.content(self.miss_branch, self.EASY_SET_SPROUT))
		self.assertIsNone(self.prq.content(sprout=self.miss_branch))
		self.assertIsNone(self.prq.content(self.EASY_SET_SPROUT, self.miss_branch))

		self.assertIsNone(self.prq.content(self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.content(self.miss_branch, self.EASY_SET_SPROUT, apparent=True))
		self.assertIsNone(self.prq.content(sprout=self.miss_branch, apparent=True))
		self.assertIsNone(self.prq.content(self.EASY_SET_SPROUT, self.miss_branch, apparent=True))


	def test_content_child(self):

		for content in (

			self.prq.content(self.pros_folder),
			self.prq.content(self.pros_folder, self.EASY_SET_SPROUT),
			self.prq.content(self.pros_folder, apparent=True),
			self.prq.content(self.pros_folder, self.EASY_SET_SPROUT, apparent=True),
			self.prq2.content(self.pros_folder),
			self.prq2.content(self.pros_folder, self.EASY_SET_SPROUT),
			self.prq2.content(self.pros_folder, apparent=True),
			self.prq2.content(self.pros_folder, self.EASY_SET_SPROUT, apparent=True),
		):
			self.assertEqual(len(content),2)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing												import Literal
from typing												import Callable
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import Transmutation
from pygwarts.magical.spells							import geminio
from pygwarts.irma.access.bookmarks						import VolumeBookmark
from pygwarts.irma.access.volume						import LibraryVolume








class AccessHandler(Transmutable):

	"""
		irma.access core super class for processing library volumes. In other words this object is to
		gather, or handle, all data during access to library. The handling of data means processing
		some strings in order to obtain information. Every AccessHandler must be declared under it's
		VolumeBookmark object, in order to be assigned to it, which in turn will be accessed by
		LibraryVolume objects. Every AccessHandler must define it's __call__ method, which must accept
		two arguments - the string to be handled, which is as line of some text, and corresponding
		LibraryVolume object as the text source. The result of lines processing must be stored in
		LibraryVolume object, so it might be further accessed by inducers.
	"""

	__call__ :Callable[[str,LibraryVolume], Literal[True] | None]

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


		if	isinstance(booklink := getattr(self, "_UPPER_LAYER", None), VolumeBookmark):

			booklink(self, "handler")
			self.loggy.debug(f"Assigned to {booklink} bookmark")
		else:
			self.loggy.info(f"AccessHandler {self} not assigned to any bookmark")




	def registered(self, volume :LibraryVolume) -> Literal[True] | None :

		"""
			Helper method that ensures LibraryVolume object "volume" either already has dictionary mapping
			for current AccessHandler object or it might be created, so returns True in both cases.
			Returns None if mapping already exist but is not a dictionary.
		"""

		match volume[self]:

			case dict()	: pass
			case None	: volume[self] = dict()
			case _:

				self.loggy.debug(f"Invalid mapping in {volume} volume")
				return
		return	True








class AccessHandlerCounter(Transmutation):

	"""
		AccessHandler utility decorator, that counts decorated AccessHandler invocations. In mutable
		chain acts as a mutation - takes decorated AccessHandler class and rewrites it's __call__
		the way with every invocation the special "access_handler_counter" field will be incremented.
		Doesn't influence decorated __call__ invocation, neither handles possible Exception.
	"""

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :
		class Handler(geminio(layer)):

			def __call__(self, line :str, volume :LibraryVolume) -> Literal[True] | None :

				self.access_handler_counter = getattr(self, "access_handler_counter", 0) +1
				self.loggy.debug(f"Handler counter incremented to {self.access_handler_counter}")

				return super().__call__(line, volume)
		return	Handler








class AccessHandlerRegisterCounter(Transmutation):

	"""
		AccessHandler utility decorator, that counts decorated AccessHandler invocations for provided
		LibraryVolume object "volume". In mutable chain acts as a mutation - takes decorated AccessHandler
		class and rewrites it's __call__ the way with every successful invocation, which is defined by
		truly return of decorated __call__, the "counter" integer mapping for current AccessHandler in
		LibraryVolume object "volume" will be incremented. Doesn't influence decorated __call__ invocation,
		neither handles possible Exception.
	"""

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :
		class Handler(geminio(layer)):

			def __call__(self, line :str, volume :LibraryVolume) -> Literal[True] | None :

				if	(handle := super().__call__(line, volume)):
					if	isinstance(vollink := volume[self], dict):


						# It is assumed decorated AccessHandler call already mapped itself in a volume,
						# so setting "counter" mapping in it may proceed. It also will serve as a guard
						# to not count if mapping not set.
						vollink.setdefault("counter",0)
						vollink["counter"] += 1
						self.loggy.debug(f"{volume} counter incremented to {vollink['counter']}")


					else:	self.loggy.debug(f"{volume} counter link was not created")
					return	handle
		return		Handler










import	os
import	unittest
from	time								import sleep
from	pygwarts.tests.hagrid				import EasySet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.planting.leafs		import LeafGrowth
from	pygwarts.hagrid.bloom.leafs			import Rejuvenation








class LeafGrowthCase(EasySet):

	"""
		GrowingPeel, Rejuvenation, regular grow
		Single sprout
		Single bough
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.EASY_HANDLER_9): os.remove(cls.EASY_HANDLER_9)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.EASY_HANDLER_9)


	def setUp(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				handler		= self.EASY_HANDLER_9
				init_level	= 10

			@GrowingPeel
			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass

		self.test_case = Sakura




	def test_growth_first_flourish(self):

		self.test_case.loggy.init_name = "growth_first_flourish"


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertFalse(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("growth_first_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_pros_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_pros_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_pros_file3}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_cons_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_cons_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_redundant_1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_first_flourish:Grown leaf \"{self.dst_redundant_2}\"",
			case_loggy.output
		)


		self.assertEqual(

			int(os.path.getmtime(self.file1)),
			int(os.path.getmtime(self.dst_file1))
		)
		self.assertEqual(

			int(os.path.getmtime(self.pros_file1)),
			int(os.path.getmtime(self.dst_pros_file1))
		)
		self.assertEqual(

			int(os.path.getmtime(self.pros_file2)),
			int(os.path.getmtime(self.dst_pros_file2))
		)
		self.assertEqual(

			int(os.path.getmtime(self.pros_file3)),
			int(os.path.getmtime(self.dst_pros_file3))
		)
		self.assertEqual(

			int(os.path.getmtime(self.cons_file1)),
			int(os.path.getmtime(self.dst_cons_file1))
		)
		self.assertEqual(

			int(os.path.getmtime(self.cons_file2)),
			int(os.path.getmtime(self.dst_cons_file2))
		)
		self.assertEqual(

			int(os.path.getmtime(self.redundant_1)),
			int(os.path.getmtime(self.dst_redundant_1))
		)
		self.assertEqual(

			int(os.path.getmtime(self.redundant_2)),
			int(os.path.getmtime(self.dst_redundant_2))
		)








	def test_growth_interval_flourish(self):

		self.test_case.loggy.init_name = "growth_interval_flourish"
		sleep(1.5)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("growth_interval_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_pros_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_pros_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_pros_file3}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_cons_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_cons_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_redundant_1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:growth_interval_flourish:Grown leaf \"{self.dst_redundant_2}\"",
			case_loggy.output
		)


		self.assertLess(

			int(os.path.getmtime(self.file1)),
			int(os.path.getmtime(self.dst_file1))
		)
		self.assertLess(

			int(os.path.getmtime(self.pros_file1)),
			int(os.path.getmtime(self.dst_pros_file1))
		)
		self.assertLess(

			int(os.path.getmtime(self.pros_file2)),
			int(os.path.getmtime(self.dst_pros_file2))
		)
		self.assertLess(

			int(os.path.getmtime(self.pros_file3)),
			int(os.path.getmtime(self.dst_pros_file3))
		)
		self.assertLess(

			int(os.path.getmtime(self.cons_file1)),
			int(os.path.getmtime(self.dst_cons_file1))
		)
		self.assertLess(

			int(os.path.getmtime(self.cons_file2)),
			int(os.path.getmtime(self.dst_cons_file2))
		)
		self.assertLess(

			int(os.path.getmtime(self.redundant_1)),
			int(os.path.getmtime(self.dst_redundant_1))
		)
		self.assertLess(

			int(os.path.getmtime(self.redundant_2)),
			int(os.path.getmtime(self.dst_redundant_2))
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)







